import re

import loopy as lp
import pymbolic.primitives as prim
import islpy as isl

from dune.codegen.generation import (get_counted_variable,
                                     get_global_context_value,
                                     )
from dune.codegen.loopy.transformations.remove_reductions import remove_all_reductions
from dune.codegen.options import get_form_option, get_option
from dune.codegen.pdelab.geometry import world_dimension
from dune.codegen.error import CodegenAutotuneError
from dune.codegen.sumfact.autotune import autotune_realization


def move_zero_assignment_up(kernel, move_up_inames):
    if len(move_up_inames) == 0:
        return kernel

    # Find the instruction we want to move around
    cond = lp.match.Tagged('set_zero')
    instructions = lp.find_instructions(kernel, cond)
    move_iname_set = set(map(lambda x: prim.Variable(x), move_up_inames))
    instr = None
    for i in instructions:
        instr_iname_set = set(i.assignee.index_tuple)
        if move_iname_set.issubset(instr_iname_set):
            # There should be only one matching instruction
            assert instr is None
            instr = i
    assert instr is not None

    # Remove it
    kernel = lp.remove_instructions(kernel, set([instr.id]))

    # Create loop domains: In order to move it upwards we need to create
    # additional loops
    iname_appendix = '_move_up'
    domains = kernel.domains
    for iname in move_up_inames:
        # Find loop bound for this iname
        for dom in domains:
            if iname in dom.get_var_names(isl.dim_type.set):
                # index = dom.get_var_names(isl.dim_type.set).index(iname)

                # TODO: Noch unklar wie man die Loop bound aus isl rausbekommt.
                todo_begin = str(dom).find(iname + ' =') + len(iname) + 3
                if todo_begin == len(iname) + 3 - 1:
                    todo_begin = str(dom).find(iname + ' <=') + len(iname) + 4
                todo_end = todo_begin + str(dom)[todo_begin:].find(' ')
                loop_bound = int(str(dom)[todo_begin:todo_end]) + 1
                break

        domain = "{{ [{0}] : 0<={0}<{1} }}".format(iname + iname_appendix, loop_bound)
        domain = lp.kernel.creation.parse_domains(domain, {})
        domains = domains + domain

    # Create tuple of correct inames for usage in  subscript below
    indices = list(instr.assignee.index_tuple)
    for i in range(len(indices)):
        if indices[i].name in move_up_inames:
            indices[i] = prim.Variable(indices[i].name + iname_appendix)
    indices = tuple(indices)

    # The new instructions needs to lie within those inames
    within_inames = []
    for i in indices:
        within_inames.append(i.name)

    # Create new instruction
    assignee = prim.Subscript(instr.assignee.aggregate, indices)
    instructions = []
    instructions.append(instr.copy(assignee=assignee,
                                   within_inames=frozenset(within_inames)))
    kernel = kernel.copy(instructions=kernel.instructions + instructions,
                         domains=domains)

    # Add dependency to inner assignment instructions
    cond = lp.match.Tagged('assignment')
    assignment_instructions = lp.find_instructions(kernel, cond)
    instr = None
    for i in assignment_instructions:
        instr_iname_set = set(i.assignee.index_tuple)
        if move_iname_set.issubset(instr_iname_set):
            # There should be only one matching instruction
            assert instr is None
            instr = i

    id_zero = instructions[0].id
    cond = lp.match.Id(instr.id)
    kernel = lp.add_dependency(kernel, cond, id_zero)

    return kernel


def _reorder_loops_in_tensor_contraction_direct(kernel, iname_order):
    """Reorder the loop nest of the tensor contractions

    iname_order is a string that specifies the loop order. We use the following convention:

    Each contraction in the sum factorization kernel has the form 'ij,jkl->kli'
    using einsum notation from numpy. iname_order should be a string like
    'iklj' if the loops should be done in order i, k, l, j.

    Without transformations those loops will be done in the order lkij.

    In the sum factorization kernel itself those inames are called:

    sf_out_inames_2_* : l
    sf_out_inames_1_* : k
    sf_out_inames_0_* : i
    red_* : j

    where * represents the current direction (0,1,2 for 3D problems).

    TODO: Maybe also support a list of the inames above?

    TODO: Different order for different direction? Could make sense when we use
    fastdg and a broadcast since the first contraction has a smaller input
    matrix.
    """
    dim = world_dimension()
    # TODO: In principle there is no need to be dimension dependent. I'm just
    # not sure how to pass the iname_order in the general case. This probably
    # needs a rework anyway so I just do the 3D case first.
    assert dim == 3

    kernel = remove_all_reductions(kernel)

    # TODO: Doc after rewrite
    reduction_iname = 'j'
    iname_dict = {'l': 'sf_out_inames_2',
                  'k': 'sf_out_inames_1',
                  'i': 'sf_out_inames_0',
                  'j': 'sf_red'}
    reduction_index = iname_order.index(reduction_iname)
    move_up_inames = list(map(lambda x: iname_dict[x], iname_order[reduction_index + 1:]))

    # cond = lp.match.Tagged('set_zero')
    cond = lp.match.Tagged('assignment')
    instructions = lp.find_instructions(kernel, cond)
    for instr in instructions:
        inames = tuple(map(lambda x: x.name, instr.assignee.index_tuple))
        current_move_up_inames = []
        for i in inames:
            for j in move_up_inames:
                if i.find(j) >= 0:
                    current_move_up_inames.append(i)

        kernel = move_zero_assignment_up(kernel, current_move_up_inames)

        # TODO: There should be a better method than searching the string for
        # 'sf_red'. Unfortunately there are sometimes Call instructions due to
        # broadcasts. That makes different ways difficult.
        regex = re.compile('sf_red_([0-9]*)')
        reduction_index = set(regex.findall(str(instr)))
        assert len(reduction_index) == 1
        reduction_index = reduction_index.pop()
        reduction_iname = 'sf_red_{}'.format(reduction_index)

        prefered_iname_order = []
        for i in inames:
            if i not in current_move_up_inames and i.find('vec') == -1:
                prefered_iname_order.append(i)
        prefered_iname_order.append(reduction_iname)
        for i in current_move_up_inames:
            prefered_iname_order.append(i)
        prefered_iname_order = tuple(prefered_iname_order)
        kernel = lp.prioritize_loops(kernel, prefered_iname_order)

    return kernel


def _get_inames_of_reduction(instr, iname_permutation):
    # Check which loops are inside the reduction loop
    reduction_index = iname_permutation[-1]
    within_reduction = tuple(i > reduction_index for i in iname_permutation)

    # Basename of inner inames
    dim = world_dimension()
    inner_inames_base = []
    for index, within in enumerate(within_reduction):
        if within:
            inner_inames_base.append('sf_out_inames_{}'.format(dim-1-index))

    # Name of inner inames and outer inames
    inner_inames = []
    outer_inames = []
    vec_inames = []
    for i in [x.name for x in instr.assignee.index_tuple]:
        for j in inner_inames_base:
            if j in i:
                inner_inames.append(i)
        if 'vec' in i:
            vec_inames.append(i)
        elif i not in inner_inames:
            outer_inames.append(i)

    # Reduction iname
    regex = re.compile('sf_red_([0-9]*)')
    reduction_index = set(regex.findall(str(instr)))
    if isinstance(instr.expression, lp.symbolic.Reduction):
        assert len(reduction_index) == 1
        reduction_index = reduction_index.pop()
        reduction_iname = 'sf_red_{}'.format(reduction_index)
    else:
        assert len(reduction_index) == 0
        reduction_iname = None

    return outer_inames, reduction_iname, inner_inames, vec_inames


def _duplicate_assignment_inames(kernel, match):
    instructions = lp.find_instructions(kernel, match)
    for instr in instructions:
        assert isinstance(instr, lp.kernel.instruction.Assignment)

        # Dependencies
        match = lp.match.Id(instr.id)
        depends_on = instr.depends_on
        depending = []
        for i in kernel.instructions:
            if instr.id in i.depends_on:
                depending.append(i.id)

        # Remove instruction
        kernel = lp.remove_instructions(kernel, set([instr.id]))

        def _duplicate_name(iname):
            iname_appendix = '_duplicate'
            return iname.name + iname_appendix

        agg_variable = kernel.temporary_variables[instr.assignee.aggregate.name]
        vectorized = isinstance(agg_variable.dim_tags[-1], lp.kernel.array.VectorArrayDimTag)
        if vectorized:
            inames = instr.assignee.index_tuple[:-1]
        else:
            inames = instr.assignee.index_tuple

        # Create new domains
        domains = kernel.domains
        new_domains = []
        for iname in inames:
            # Find loop bound for the corresponding domain
            for dom in domains:
                if iname.name in dom.get_var_names(isl.dim_type.set):
                    # TODO There must be better way to get this information using isl
                    str_begin = str(dom).find(iname.name + ' =') + len(iname.name) + 3
                    if str_begin == len(iname.name) + 3 - 1:
                        str_begin = str(dom).find(iname.name + ' <=') + len(iname.name) + 4
                    str_end = str_begin + str(dom)[str_begin:].find(' ')
                    loop_bound = int(str(dom)[str_begin:str_end]) + 1
                    break

            # Create new domain
            domain = "{{ [{0}] : 0<={0}<{1} }}".format(_duplicate_name(iname), loop_bound)
            domain = lp.kernel.creation.parse_domains(domain, {})
            new_domains.append(domain)
        for domain in new_domains:
            domains = domains + domain

        # Create new inames
        new_inames = tuple(prim.Variable(_duplicate_name(i)) for i in inames)
        if vectorized:
            new_inames = new_inames + (instr.assignee.index_tuple[-1],)

        # Create new instruction within the new inames
        assignee = prim.Subscript(instr.assignee.aggregate, new_inames)
        new_instruction = instr.copy(assignee=assignee,
                                     depends_on=depends_on,
                                     within_inames=frozenset([i.name for i in new_inames]))
        kernel = kernel.copy(instructions=kernel.instructions + [new_instruction],
                             domains=domains)

        # Restore dependencies
        for dep in depending:
            match = lp.match.Id(dep)
            kernel = lp.add_dependency(kernel, match, new_instruction.id)

    return kernel


def _reorder_loops_in_tensor_contraction_accum(kernel, iname_permutation):
    dim = world_dimension()

    # Nothing to do if permutation is identity
    if iname_permutation == tuple(range(dim+1)):
        return kernel

    # Use names used in sum factorization kernel (without the index that distinguishes the different directions)
    default_iname_order = ['sf_out_inames_{}'.format(dim-1-i) for i in range(dim)] + ['sf_red']
    from dune.codegen.sumfact.permutation import permute_backward
    new_iname_order = permute_backward(default_iname_order, iname_permutation)

    # Get the real names with direction indices in the right order
    def _current_new_iname_order(outer, reduction, inner, new_iname_order):
        if reduction:
            reduction = [reduction]
        else:
            reduction = []
        all_inames = outer + reduction + inner
        current_iname_order = []
        for i in new_iname_order:
            for j in all_inames:
                if i in j:
                    current_iname_order.append(j)
        return current_iname_order

    for instr in kernel.instructions:
        # We can directly use lp.prioritize_loops if:
        # - The reduction is the innermost loop
        # - There is no reduction (eg reduced direction on faces)
        if iname_permutation[-1] == dim or not isinstance(instr.expression, lp.symbolic.Reduction):
            # Inames used in this reduction
            outer_inames, reduction_iname, inner_inames, vec_inames = _get_inames_of_reduction(instr,
                                                                                               iname_permutation)
            current_iname_order = _current_new_iname_order(outer_inames,
                                                           reduction_iname,
                                                           inner_inames,
                                                           new_iname_order)
            kernel = lp.prioritize_loops(kernel, tuple(current_iname_order))
            continue

        assert isinstance(instr.expression.operation, lp.library.reduction.SumReductionOperation)

        # Dependencies. Note: We change the instructions below so we cannot do
        # depends_on=instr.depends_on. But we know that the id of instr is
        # still valid so we can use that!
        match = lp.match.Id(instr.id)
        depends_on = lp.find_instructions(kernel, match)[0].depends_on
        depending = []
        for i in kernel.instructions:
            if instr.id in i.depends_on:
                depending.append(i.id)

        # Remove reduction
        kernel = lp.remove_instructions(kernel, set([instr.id]))

        # Inames used in this reduction
        outer_inames, reduction_iname, inner_inames, vec_inames = _get_inames_of_reduction(instr, iname_permutation)

        # Create dim_tags
        dim_tags = ','.join(['f'] * len(inner_inames))
        agg_variable = kernel.temporary_variables[instr.assignee.aggregate.name]
        vectorized = isinstance(agg_variable.dim_tags[-1], lp.kernel.array.VectorArrayDimTag)
        if vectorized:
            assert len(vec_inames) == 1
            dim_tags = dim_tags + ',vec'

        # Create shape
        agg_inames = tuple(i.name for i in instr.assignee.index_tuple)
        agg_shape = agg_variable.shape
        shape = tuple(agg_shape[agg_inames.index(i)] for i in inner_inames)
        if vectorized:
            shape = shape + (agg_shape[-1],)

        # Update temporary_variables of this kernel
        from dune.codegen.loopy.temporary import DuneTemporaryVariable
        accum_variable = get_counted_variable('accum_variable')
        var = {accum_variable: DuneTemporaryVariable(accum_variable,
                                                     dtype=agg_variable.dtype,
                                                     shape=shape,
                                                     dim_tags=dim_tags,
                                                     managed=True)}
        kernel.temporary_variables.update(var)

        # Set accumulation variable to zero
        accum_init_inames = tuple(prim.Variable(i) for i in inner_inames)
        if vectorized:
            accum_init_inames = accum_init_inames + (prim.Variable(vec_inames[0]),)
        assignee = prim.Subscript(prim.Variable(accum_variable,), accum_init_inames)
        accum_init_id = instr.id + '_accum_init'
        accum_init_instr = lp.Assignment(assignee,
                                         0,
                                         within_inames=instr.within_inames,
                                         id=accum_init_id,
                                         depends_on=depends_on,
                                         tags=('accum_init',),
                                         )
        kernel = kernel.copy(instructions=kernel.instructions + [accum_init_instr,])

        # Accumulate in temporary variable
        assignee = prim.Subscript(prim.Variable(accum_variable,), accum_init_inames)
        expression = prim.Sum((assignee, instr.expression.expr))
        within_inames = frozenset(tuple(instr.within_inames) + instr.expression.inames)
        accum_id = instr.id + '_accum'
        accum_instr = lp.Assignment(assignee,
                                    expression,
                                    within_inames=within_inames,
                                    id=accum_id,
                                    depends_on=frozenset([accum_init_id,]),
                                    tags=('accum',),
                                    )
        kernel = kernel.copy(instructions=kernel.instructions + [accum_instr,])

        # Assign accumulation result
        assign_id = instr.id + '_assign'
        assign_instr = lp.Assignment(instr.assignee,
                                     assignee,
                                     within_inames=within_inames,
                                     id=assign_id,
                                     depends_on=frozenset([accum_id,]),
                                     tags=('accum_assign',),
                                     )
        kernel = kernel.copy(instructions=kernel.instructions + [assign_instr,])

        # Restore dependencies
        for dep in depending:
            match = lp.match.Id(dep)
            kernel = lp.add_dependency(kernel, match, assign_id)

        # Reordering loops only works if we duplicate some inames
        duplicate_inames = tuple(i.name for i in accum_init_inames)
        if vectorized:
            duplicate_inames = duplicate_inames[:-1]
        match = lp.match.Id(accum_init_id)
        kernel = lp.duplicate_inames(kernel, duplicate_inames, match)
        match = lp.match.Id(assign_id)
        kernel = lp.duplicate_inames(kernel, duplicate_inames, match)

        # Change loop order
        current_iname_order = _current_new_iname_order(outer_inames,
                                                       reduction_iname,
                                                       inner_inames,
                                                       new_iname_order)
        kernel = lp.prioritize_loops(kernel, tuple(current_iname_order))

    return kernel


def reorder_loops_in_tensor_contraction(kernel, iname_permutation, accum_variable=True):
    """Change loop order in tensor contractions of sum factorization kernel

    Since there is a reduction involved this implies more than just reordering
    loops. There are two possibilities to handle the reduction:

    1) Generate an object of correct size for the accumulation and assign to
    the result data structure afterwards

    2) Make sure to set the correct entries of the result data structure to
    zero and accumulate inplace

    Parameters
    ----------
    kernel: loopy.kernel.LoopKernel
    iname_permutation: tuple of ints
        How to permute the loop inames. Should contain some permutation of the
        numbers 0 to dim+1
    accum_variable: bool
        If True use method 1 else use method 2 from above

    Notes
    -----
    Using einsum notation from numpy a contraction of a sum factorization
    kernel can be written as:

    - 3D: 'ij,jkl->kli'
    - 2D: 'ij,jk->ki'

    In the sum factorization kernel itself those inames are called:

    sf_out_inames_2_* : l
    sf_out_inames_1_* : k
    sf_out_inames_0_* : i
    sf_red_* : j

    where * represents the current direction (0,1,2 for 3D problems). The
    default order of the loops is

    (sf_out_iname_2_*, sf_out_iname_1_*, sf_out_iname_0_*, red_*)

    A permutation vector of (3,2,0,1) would mean that we do

    (sf_out_iname_0_*, red_*, sf_out_iname_1_*, sf_out_iname_2_*)

    instead.

    """
    if accum_variable:
        return _reorder_loops_in_tensor_contraction_accum(kernel, iname_permutation)
    else:
        # Need to adapts this!
        assert False
        return _reorder_loops_in_tensor_contraction_direct(kernel, iname_permutation)


def tensor_contraction_loop_order_generator(kernel):
    dim = world_dimension()
    assert dim == 3

    indices = ['l', 'k', 'i', 'j']
    import itertools
    for loop_order in itertools.permutations(indices):
        # palpo TODO: Heavy culling for 'quick' tests during development
        if loop_order[0] != 'l' or loop_order[1] != 'k':
            continue

        order = ''.join(loop_order)
        new_kernel = reorder_loops_in_tensor_contraction(kernel, order, True)
        yield new_kernel, ['reorder_loops_in_tensor_contraction_{}_True'.format(order),]

        new_kernel = reorder_loops_in_tensor_contraction(kernel, loop_order, False)
        yield new_kernel, ['reorder_loops_in_tensor_contraction_{}_False'.format(order),]


def simple_autotuner(kernel_generator, signature):
    # palpo TODO
    from dune.codegen.options import set_option
    set_option("autotune_google_benchmark", True)

    kernel, transformations = next(kernel_generator)
    best_cost = autotune_realization(kernel=kernel, signature=signature, transformations=transformations)
    best_kernel = kernel
    for kernel, transformations in kernel_generator:
        cost = autotune_realization(kernel=kernel, signature=signature, transformations=transformations)
        if cost < best_cost:
            best_cost = cost
            best_kernel = kernel
    return best_kernel


def autotune_tensor_contraction_loop_order(kernel, signature):
    from dune.codegen.loopy.transformations.matchfma import match_fused_multiply_add
    kernel = match_fused_multiply_add(kernel)

    generator = tensor_contraction_loop_order_generator(kernel)
    return simple_autotuner(generator, signature)


def sumfact_performance_transformations(kernel, signature):
    if kernel.name.startswith('sfimpl'):
        # 2D
        # kernel = reorder_loops_in_tensor_contraction(kernel, (2,0,1), True)

        # 3D
        # kernel = reorder_loops_in_tensor_contraction(kernel, (3,2,0,1), True)
        # kernel = reorder_loops_in_tensor_contraction(kernel, (1,2,0,3), True)

        # kernel = autotune_tensor_contraction_loop_order(kernel, signature)

        pass
    return kernel
